#include "ethernet.h"

/* Type definitions */
typedef struct {
    uint32_t DES0;
    uint32_t DES1;
    uint32_t DES2;
    uint32_t DES3;
} TXRX_descriptor_T;

volatile TXRX_descriptor_T RX_descriptor_list[1];
volatile uint8_t rx_buffer[N_PACKET_BYTES];
volatile TXRX_descriptor_T TX_descriptor_list[2];
volatile uint8_t tx_buffer[2][N_PACKET_BYTES];
volatile uint8_t *tx_buffer_ptr;

volatile uint32_t N_requested_frames;

void init_ethernet()
{
    N_requested_frames = 0;
    
    tx_buffer_ptr = &(tx_buffer[0][0]);    // assign global pointer so that it can be used externally
    
    // initialize the pins (alternate function...)
    setAlternateFunction(PORTG, 11, 11);    // RMII_TX_EN   = PG11, SB183
    setAlternateFunction(PORTG, 13, 11);    // RMII_TXD0    = PG13, SB182
    setAlternateFunction(PORTB, 13, 11);    // RMII_TXD1    = PB13, JP7
    setAlternateFunction(PORTC, 4,  11);    // RMII_RXD0    = PC4,  SB178
    setAlternateFunction(PORTC, 5,  11);    // RMII_RXD1    = PC5,  SB181
    setAlternateFunction(PORTA, 7,  11);    // RMII_CRS_DV  = PA7,  JP6
    setAlternateFunction(PORTC, 1,  11);    // RMII_MDC     = PC1,  SB164
    setAlternateFunction(PORTA, 2,  11);    // RMII_MDIO    = PA2,  SB160
    setAlternateFunction(PORTA, 1,  11);    // RMII_REF_CLK = PA1,  SB13
    
    // enable ethernet interrupt (NVIC level) (interrupt #61 which is bit 29 in ISER1 since 61 - 32 = 29)
    NVIC->ISER[1] |= (1 << 29);
    
    // use RMII. Must be done before enabling MAC clocks and before enabling the MAC itself.
    RCC->APB2ENR |= (1 << 14);    // enable clock for SYSCFG-block
    SYSCFG->PMC |= (1 << 23);     // use RMII
    
    // enable clock for ethernet and DMA1, DMA2. TODO: is it really required to enable DMA1 and DMA2 clock?
    RCC->AHB1ENR |= (1 << 21) | (1 << 22) | (1 << 25) | (1 << 26) | (1 << 27) | (1 << 28);
    
    // setup DMA (RM0410, 42.6.1, page 1808)
    // 1.
    ETH->DMABMR |= (1 << 0);           // issue software reset. Very important!
    while((ETH->DMABMR >> 0) == 1);    // wait for software reset bit = 0 as this signals that the software reset is done
    
    // set MAC address = 11:22:33:44:55:66. Be aware of byte-order!
    ETH->MACA0HR = 0xFFFF & 0x6655;
    ETH->MACA0LR = 0x44332211;
    
    // 2.
    ETH->DMAIER |= (1 << 16) | (1 << 15) | (1 << 6);     // normal interrupt enabled, abnormal interrupt enabled, receive interrupt enabled
    
    // 3.
    TX_descriptor_list[0].DES0 &= ~(1 << 31);    // OWN = 0 ==> descriptor is owned by CPU. Do not transmit yet!
    TX_descriptor_list[0].DES0 |= (1 << 29);     // LS = 1 ==> buffer contains last segment of the frame
    TX_descriptor_list[0].DES0 |= (1 << 28);     // FS = 1 ==> buffer contains first segment of the frame
    TX_descriptor_list[0].DES0 |= (1 << 20);     // TCH = 1 ==> second address chained
    //TX_descriptor_list[0].DES0 |= (1 << 21);     // TER = 1 ==> end of ring
    TX_descriptor_list[0].DES1 |= (((N_PACKET_BYTES) & 0x1FFF) << 0);    // buffer 1 size = N_PACKET_BYTES bytes
    TX_descriptor_list[0].DES2 = (uint32_t)(&(tx_buffer[0]));
    TX_descriptor_list[0].DES3 = (uint32_t)(&(TX_descriptor_list[1]));
    
    TX_descriptor_list[1].DES0 &= ~(1 << 31);    // OWN = 0 ==> descriptor is owned by CPU. Do not transmit yet!
    TX_descriptor_list[1].DES0 |= (1 << 29);     // LS = 1 ==> buffer contains last segment of the frame
    TX_descriptor_list[1].DES0 |= (1 << 28);     // FS = 1 ==> buffer contains first segment of the frame
    TX_descriptor_list[1].DES0 |= (1 << 21);     // TER = 1 ==> end of ring
    TX_descriptor_list[1].DES1 |= (((N_PACKET_BYTES) & 0x1FFF) << 0);    // buffer 1 size = N_PACKET_BYTES bytes
    TX_descriptor_list[1].DES2 = (uint32_t)(&(tx_buffer[1]));
    ETH->DMATDLAR = (uint32_t)(&(TX_descriptor_list[0]));
    
    RX_descriptor_list[0].DES1 &= ~(1 << 31);   // DIC = 0 ==> receive interrupt not disabled
    RX_descriptor_list[0].DES1 |= (1 << 15);    // RER = 1 ==> end of ring
    RX_descriptor_list[0].DES1 |= ((0x1FFF & N_PACKET_BYTES) << 0);    // buffer 1 size = N_PACKET_BYTES bytes
    RX_descriptor_list[0].DES2 = (uint32_t)(&rx_buffer);
    ETH->DMARDLAR = (uint32_t)(&RX_descriptor_list);
    RX_descriptor_list[0].DES0 |= (1 << 31);    // OWN = 1 ==> descriptor is owned by DMA. Ready to receive!
    
    // 4.
    // set MAC frame filter settings
    //ETH->MACFFR |= (1 << 31) | (1 << 4);   // RA  = 1 ==> Receive all, PAM = 1 ==> Pass all multicast. For debug!
    //ETH->MACFFR |= (1 << 5) | (0x3 << 6);    // filter out all broadcast frames and pass control frames (that passed the filter!) to the application. For debug!
    ETH->MACFFR |= (1 << 5);    // filter out all broadcast frames
    
    // 5.
    // write phy registers to set up the phy
    phy_write((1 << 12), 0);                    // enable Auto-Negotiation
    while(((phy_read(1) >> 5) & 0x01) == 0);    // wait for Auto-Negotiation to be done
    
    ETH->MACCR |= (1 << 14) | (1 << 11);    // FES = 1 ==> 100Mbit/s, DM  = 1 ==> Fullduplex. This should be adjusted to fit the actual PHY config! TODO!
    wait_cycles(1000000);    // According to some errata sheet, we need to wait for some time after writing to a specific register, before we can write to the very same register again, otherwise we could have corrupted data in the register.
    
    // 5.
    ETH->MACCR |= (1 << 3) | (1 << 2);     // TE = 1 ==> Transmitter enabled, RE = 1 ==> Receiver enabled
    
    // 6.
    // receive store and forward, transmit store and forward, start transmission, start receive
    ETH->DMAOMR |= (1 << 25) | (1 << 21) | (1 << 13) | (1 << 1);
}

void phy_write(uint32_t data, uint32_t phy_register)
{
    while(((ETH->MACMIIAR >> 0) & 0x01) == 1);    // wait until the MII is not busy anymore
    
    ETH->MACMIIDR = 0xFFFF & data;    // set data should be written to the phy
    
    // set PHY_ADDRESS and clock range aswell as the 'MII write' bit. Also here we select to which phy register we want to write to
    ETH->MACMIIAR = ((0x1F & PHY_ADDRESS) << 11) | ((0x1F & phy_register) << 6) | (0x4 << 2) | (1 << 1);
    
    ETH->MACMIIAR |= (1 << 0) | (1 << 1);    // issue the write operation
    
    while(((ETH->MACMIIAR >> 0) & 0x01) == 1);    // wait until the MII is not busy anymore
}

uint32_t phy_read(uint32_t phy_register)
{
    while(((ETH->MACMIIAR >> 0) & 0x01) == 1);    // wait until the MII is not busy anymore
    
    // set PHY_ADDRESS and clock range. Also here we select to which phy register we want to read from
    ETH->MACMIIAR = ((0x1F & PHY_ADDRESS) << 11) | ((0x1F & phy_register) << 6) | (0x4 << 2);
    
    ETH->MACMIIAR |= (1 << 0);    // issue the write operation
    
    while(((ETH->MACMIIAR >> 0) & 0x01) == 1);    // wait until the MII is not busy anymore
    
    return ((ETH->MACMIIDR) & 0xFFFF);    // get data that is read from the phy
}

void transmit_ethernet_frame(uint32_t descriptor_index)
{
    // destination mac
    tx_buffer[descriptor_index][0] = 0xCA;
    tx_buffer[descriptor_index][1] = 0xFF;
    tx_buffer[descriptor_index][2] = 0xEE;
    tx_buffer[descriptor_index][3] = 0xBA;
    tx_buffer[descriptor_index][4] = 0xBE;
    tx_buffer[descriptor_index][5] = 0x12;
    
    // source mac
    tx_buffer[descriptor_index][6] = 0x11;
    tx_buffer[descriptor_index][7] = 0x22;
    tx_buffer[descriptor_index][8] = 0x33;
    tx_buffer[descriptor_index][9] = 0x44;
    tx_buffer[descriptor_index][10] = 0x55;
    tx_buffer[descriptor_index][11] = 0x66;
    
    // EtherType: https://en.wikipedia.org/wiki/EtherType#Examples
    tx_buffer[descriptor_index][12] = 0x08;
    tx_buffer[descriptor_index][13] = 0x00;
    
    while(((TX_descriptor_list[descriptor_index].DES0 >> 31) & 0x01) == 1);    // wait for previous transmission to be done
    
    TX_descriptor_list[descriptor_index].DES0 |= (1 << 31);    // OWN = 1 ==> descriptor is now owned by DMA. Transmit now!
    ETH->DMATPDR = 0x12345678;    // "wake up" the DMA so that it will transmit the packet
    
    //while(((TX_descriptor_list[descriptor_index].DES0 >> 31) & 0x01) == 1);    // wait for transmission to be done
}

void ETH_IRQHandler()
{
    if(((ETH->DMASR >> 6) & 0x01) == 1)
    {
        ETH->DMASR |= (1 << 6);    // clear RS bit
        
        /*
        // print received frame
        // receiver's MAC
        send_string_UART("receiver's MAC = ");
        send_byte_hex_UART(rx_buffer[0]);
        send_string_UART(":");
        send_byte_hex_UART(rx_buffer[1]);
        send_string_UART(":");
        send_byte_hex_UART(rx_buffer[2]);
        send_string_UART(":");
        send_byte_hex_UART(rx_buffer[3]);
        send_string_UART(":");
        send_byte_hex_UART(rx_buffer[4]);
        send_string_UART(":");
        send_byte_hex_UART(rx_buffer[5]);
        send_string_UART("\n");
        
        // sender's MAC
        send_string_UART("senders's MAC = ");
        send_byte_hex_UART(rx_buffer[6]);
        send_string_UART(":");
        send_byte_hex_UART(rx_buffer[7]);
        send_string_UART(":");
        send_byte_hex_UART(rx_buffer[8]);
        send_string_UART(":");
        send_byte_hex_UART(rx_buffer[9]);
        send_string_UART(":");
        send_byte_hex_UART(rx_buffer[10]);
        send_string_UART(":");
        send_byte_hex_UART(rx_buffer[11]);
        send_string_UART("\n");
        
        // EtherType
        // https://en.wikipedia.org/wiki/EtherType#Examples
        send_string_UART("EtherType = 0x");
        send_byte_hex_UART(rx_buffer[12]);
        send_byte_hex_UART(rx_buffer[13]);
        send_string_UART("\n");
        
        // data (payload)
        send_string_UART("data = ");
        send_string_UART((char*)&(rx_buffer[14]));
        send_string_UART("\n");
        */
        
        if((rx_buffer[14] == 'r') && (rx_buffer[15] == 'u') && (rx_buffer[16] == 'n'))
        {
            N_requested_frames = (((uint32_t)(rx_buffer[17])) << 24)
                                | (((uint32_t)(rx_buffer[18])) << 16)
                                | (((uint32_t)(rx_buffer[19])) << 8)
                                | (((uint32_t)(rx_buffer[20])) << 0);
        }
        
        // give descriptor ownership back to DMA so that it can be used for the next frame
        RX_descriptor_list[0].DES0 |= (1 << 31);    // OWN = 1 ==> descriptor is owned by DMA
    }
}
